import datetime
from unittest import skip
from freezegun import freeze_time
from dateutil.relativedelta import relativedelta

from odoo import Command
from odoo.tests import HttpCase, tagged, new_test_user
from odoo.tools import get_timedelta
from odoo.exceptions import UserError
from odoo.addons.openerp_enterprise.tests.common import generate_uuid
from odoo.addons.openerp_enterprise.tests.test_subscription import SubscriptionCommon

class MaintenanceCommon(SubscriptionCommon):
    def init_data(self):
        self.maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        self.maintenance_product.product_subscription_pricing_ids = [
            Command.clear(),
            Command.create({'plan_id': self.env.ref('openerp_enterprise.subscription_plan_odoo_monthly_plan').id, 'price': 16.0}),
            Command.create({'plan_id': self.env.ref('openerp_enterprise.subscription_plan_odoo_yearly_plan').id, 'price': 192.0}),
        ]

        super().init_data()
        self.maintenance_subscription = self.main_subscription._get_maintenance_subscription()
        self.maintenance_line = self.maintenance_subscription._get_maintenance_line()

    def _setup_separate_maintenance_sub(self, partner, quotation_template):
        """Create a distinct maintenance subscription to test used bases before saas-15.3."""
        order = self._create_so_from_template(partner, quotation_template)
        order.order_line.filtered(lambda l: l.product_id != self.maintenance_product).unlink()
        self.main_subscription.maintenance_subscription_id = order
        self.maintenance_subscription = self.main_subscription._get_maintenance_subscription()
        self.maintenance_line = self.maintenance_subscription._get_maintenance_line()

class MaintenanceReferrer(MaintenanceCommon):

    @classmethod
    def init_master_data(cls):
        super().init_master_data()
        grade = cls.env.ref('openerp_portal.res_partner_grade_learning')
        grade.default_commission_plan_id = cls.env['commission.plan'].create({
            'name': "Commission Plan"
        })
        cls.referrer = cls.env['res.partner'].create({
            'name': 'Subscription Referrer',
            'country_id': cls.env.ref('base.be').id,
            'property_product_pricelist': cls.env.ref('openerp_enterprise.pl22_eu_high').id,
            'grade_id': grade.id,
            'commission_plan_id': grade.default_commission_plan_id.id,
            'email': 'referrer@test.com',
        })
        cls.referrer_user = new_test_user(cls.env,
            login="learning-partner@example.org", groups="base.group_portal",
            partner_id=cls.referrer.id,
        )

    def init_so(self):
        super().init_so()
        self.so.referrer_id = self.referrer


class MaintenanceMonthlyTest(MaintenanceCommon):
    quotation_template = MaintenanceCommon.qt_custom_monthly

    def test_first_ping(self):
        """ Test ping update database loc count and monthly subscription qty """
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
        }

        self.send_ping(db_uuid, modules)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        self.assertEqual(len(database), 1, "You should find a database")
        self.assertEqual(database.maintenance_lines_count, 1000, "The extra modules line count should be 1000 line")
        database.subscription_id = self.main_subscription
        self.assertEqual(self.maintenance_line.product_uom_qty, 10.0, "Should be ten 100/loc per month")
        self.assertEqual(self.maintenance_subscription.subscription_state, '3_progress',
            "Monthly Maintenance sub should be directly in progress")

    def test_renewal_payment_link(self):
        uom_100loc_month = self.env.ref('openerp_enterprise.product_uom_maintenance_100loc')
        db_uuid = generate_uuid()
        modules = {'module_1': 300}

        # We receive a ping with custom LOC: maintenance should be updated for monthly
        self.send_ping(db_uuid, modules, code=self.code)
        mnt_line_torenew = self.main_subscription._get_maintenance_line()
        self.assertEqual(mnt_line_torenew.product_uom_qty, 3.0)
        self.assertEqual(mnt_line_torenew.product_uom, uom_100loc_month)

        # Create a renewal, ensure the maintenance qty are kept.
        self.main_subscription._create_invoices()._post()
        with freeze_time(self.main_subscription.next_invoice_date):
            res = self.main_subscription.prepare_renewal_order()
            renewal_order = self.env['sale.order'].browse(res['res_id'])
            self.assertEqual(renewal_order.subscription_state, '2_renewal')
            self.assertFalse(renewal_order.to_renew, "Draft renewal should not be considered as 'To Renew'")
        mnt_line_renew = renewal_order._get_maintenance_line()
        self.assertEqual(len(mnt_line_renew), 1, "The maintenance line should be kept on renewal order")
        self.assertEqual(mnt_line_renew.product_uom_qty, 3.0)
        self.assertEqual(mnt_line_renew.product_uom, uom_100loc_month)
        self.assertEqual(renewal_order.amount_untaxed, self.main_subscription.amount_untaxed,
                         "The renewal should have the same amount as current subscription")

        # Generate a payment link so that customer that pay the renew
        # immediately and ensure we propose the right default amount
        ctx = {"active_model": "sale.order", "active_id": renewal_order.id, "active_ids": renewal_order.ids}
        defaults = self.env["payment.link.wizard"].with_context(**ctx).default_get(['amount'])
        self.assertEqual(mnt_line_renew.product_uom_qty, 3.0,
                         "The maintenance quantity should not have changed")
        self.assertEqual(renewal_order.amount_untaxed, self.main_subscription.amount_untaxed,
                         "The renewal should not have changed")
        self.assertEqual(defaults['amount'], self.main_subscription.amount_total)


    def test_renewal_monthly_to_yearly(self):
        """Renew the main subscription to switch to yearly

        - the monthly subscription should be closed
        - the new yearly subscription should be live and be assigned of the database
        """
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
        }

        self.send_ping(db_uuid, modules, code=self.code)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        self.assertEqual(database.subscription_id, self.main_subscription,
            "Database should be linked to Main Subscription"
        )
        self.assertEqual(self.maintenance_line.product_uom_qty, 10.0, "Should be ten 100/loc per month")
        self.assertEqual(self.maintenance_line.price_subtotal, 160, "Total should be 160")
        self.assertEqual(
            self.main_subscription.plan_id.billing_period,
            get_timedelta(1, "month"),
            "Subscription should start as monthly"
        )
        invoice = self.main_subscription._create_invoices()
        invoice._post()

        res = self.main_subscription.prepare_renewal_order()
        renewal_order = self.env['sale.order'].browse(res['res_id'])
        renewal_order.sale_order_template_id = self.env.ref('openerp_enterprise.custom_template_odoo_yearly')
        renewal_order.order_line._compute_product_uom()
        renewal_order.action_confirm()
        self.env.cr.precommit.run()
        renewal_maintenance_line = renewal_order._get_maintenance_line()
        self.assertEqual(renewal_maintenance_line.price_subtotal, 1920, "Total should be 1920 (160 x 12)")
        self.assertEqual(
            renewal_order.plan_id.billing_period,
            get_timedelta(1, "year"),
            "Subscription change to yearly"
        )
        self.assertEqual(database.subscription_id, renewal_order,
            "Database should now be linked to Renewal Subscription"
        )


@tagged('-at_install', 'post_install')
class MaintenanceMonthlyReferrerTest(MaintenanceReferrer, HttpCase):
    quotation_template = MaintenanceReferrer.qt_custom_monthly

    @skip('disabled until message_class passthrough is fixed in account_avatax_sale_subscription')
    def test_activate_maintenance(self):
        db_uuid = generate_uuid()
        modules = {'module_1': 1000}

        # ensure Maintenance is not activated
        self.main_subscription._get_maintenance_line().unlink()
        self.assertFalse(self.main_subscription.maintenance_is_activated)

        # force sub. code for tour and ping with updated custom modules LOC
        self.main_subscription.client_order_ref = "M2023REFERRED"
        self.send_ping(db_uuid, modules, code="M2023REFERRED")
        self.assertTrue(self.main_subscription.database_ids)
        self.assertEqual(self.main_subscription._get_maintenance_loc_count(), 1000)
        self.main_subscription._create_invoices()._post()

        # start tour to activate the maintenance
        self.start_tour('/my/subscription', 'partner_activate_maintenance_monthly', login=self.referrer_user.login)
        self.assertTrue(self.main_subscription.maintenance_is_activated)

    def test_one_subscription(self):
        self.assertEqual(self.main_subscription._get_maintenance_subscription(), self.maintenance_subscription,
            "Main and maintenance subscription should be the same")
        self.assertEqual(self.main_subscription.referrer_id, self.referrer, "Referrer should be on the subscription")
        self.assertTrue(self.main_subscription._get_maintenance_line(),
            "Main Subscription should have a maintenance line")

    def test_renewal_monthly_to_yearly_referrer(self):
        """ Renew the main subscription to switch to yearly """
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
        }
        self.send_ping(db_uuid, modules, code=self.code)

        invoice = self.main_subscription._create_invoices()
        invoice._post()
        res = self.main_subscription.prepare_renewal_order()
        renewal_order = self.env['sale.order'].browse(res['res_id'])
        renewal_order.sale_order_template_id = self.env.ref('openerp_enterprise.custom_template_odoo_yearly')
        renewal_order.order_line._compute_product_uom()
        renewal_order.action_confirm()
        self.assertEqual(
            self.main_subscription.plan_id.billing_period,
            get_timedelta(1, "month"),
            "Renewed Subscription should stay as monthly"
        )
        self.assertEqual(
            renewal_order.plan_id.billing_period,
            get_timedelta(1, "year"),
            "Subscription change to yearly"
        )
        self.assertEqual(renewal_order._get_maintenance_line().price_subtotal, 160*12,
            "Maintenance should be 12 times bigger")


class MaintenanceYearlyTest(MaintenanceCommon):
    quotation_template = MaintenanceCommon.qt_custom_yearly

    def test_yearly_mnt_in_progress(self):
        self.assertEqual(self.main_subscription._get_maintenance_line().product_uom_qty, 0,
            "Sold quantity should start at 0")
        self.assertEqual(self.maintenance_subscription, self.main_subscription,
            "Yearly Maintenance sub should be the same as the main_subscription")
        self.assertEqual(self.maintenance_subscription.subscription_state, '3_progress',
            "Yearly Maintenance sub with no referrer should be directly in progress")

    def test_upsell(self):
        """ Upsell of new loc """

        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
            'module_2': 1000,
        }
        self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(self.maintenance_subscription, self.main_subscription,
            "Yearly Maintenance sub should be the same as the main_subscription")
        self.assertTrue(self.maintenance_subscription.to_renew, "Should need to upsell")
        self.main_subscription._create_invoices()
        self.main_subscription.invoice_ids._post()
        # To renew should be True
        # Trigger Upsell sale order
        upsell_so = self.upsell(self.maintenance_subscription)
        # Confirm SO
        upsell_so.action_confirm()
        self.assertFalse(self.maintenance_subscription.to_renew, "Should not need to upsell anymore")
        modules['module_1'] = 2500
        self.send_ping(db_uuid, modules, code=self.code)
        self.assertFalse(self.maintenance_subscription.to_renew,
            "Should not trigger an upsell since the last upsell SO is less then 30 days ago")

        upsell_so.date_order = datetime.date.today() - datetime.timedelta(days=40)
        self.send_ping(db_uuid, modules, code=self.code)
        self.assertTrue(self.maintenance_subscription.to_renew,
            "Should trigger an upsell since the last upsell is more than 30d days ago")

    def test_cancel_upsell(self):
        """ Test Cancel Upsell order """
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
            'module_2': 1000,
        }
        #Set initial value of sold maintenance
        self.main_subscription._get_maintenance_line().product_uom_qty = 20
        self.main_subscription._create_invoices()._post()

        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal', "No need to upsell at this stage")
        modules['module_1'] = 1500
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'upsell', "need to upsell at this stage")
        # Trigger Upsell sale order
        upsell_so = self.upsell(self.main_subscription)
        self.assertEqual(upsell_so._get_maintenance_line().product_uom_qty, 5.0, "Should upsell 500 lines of codes")
        modules['module_1'] = 1000
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal', "No need to upsell at this stage")

    def test_renewal_yearly(self):
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
            'module_2': 1000,
        }
        # Set initial value of sold maintenance
        self.main_subscription._get_maintenance_line().product_uom_qty = 20
        self.main_subscription._create_invoices()._post()
        res = self.send_ping(db_uuid, modules, code=self.code)
        current_next_invoice = self.main_subscription.next_invoice_date

        res = self.main_subscription.prepare_renewal_order()
        order = self.env['sale.order'].browse(res['res_id'])
        order.action_confirm()
        self.assertFalse(self.main_subscription.maintenance_subscription_id, "No maintenance subscription should be created")
        self.assertEqual(self.main_subscription.state, "sale")
        self.assertTrue(self.main_subscription.locked)
        self.assertTrue(self.main_subscription.subscription_state, "5_renewed")
        self.assertEqual(
            self.main_subscription.next_invoice_date, current_next_invoice,
            "Upon renew, the renewal should take the current Next Invoice Date",
        )

    def test_valid_date(self):
        """ Test different use case that change the valid date of the database """
        self.main_subscription._create_invoices()._post()
        db_uuid = generate_uuid()
        modules = {'module_1': 1000}
        self.send_ping(db_uuid, modules, code=self.code)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), datetime.date.today() + relativedelta(months=1, days=1),
            "The maintenance subscription need an upsell, the valid date should be today + 1 month and 1 day")
        self.upsell(self.maintenance_subscription).action_confirm()
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), self.main_subscription.next_invoice_date + datetime.timedelta(days=1),
            "The maintenance subscription is valid and paid, the valid date should be the end of the main subscription + 1 days")


@tagged('-at_install', 'post_install')
class MaintenanceYearlyReferrerTest(MaintenanceReferrer, HttpCase):
    quotation_template = MaintenanceReferrer.qt_custom_yearly

    def test_activate_maintenance(self):
        db_uuid = generate_uuid()
        modules = {'module_1': 1000}

        # ensure Maintenance is not activated
        self.main_subscription._get_maintenance_line().unlink()
        self.assertFalse(self.main_subscription.maintenance_is_activated)

        # force sub. code for tour and ping with updated custom modules LOC
        self.main_subscription.client_order_ref = "M2023REFERRED"
        self.send_ping(db_uuid, modules, code="M2023REFERRED")
        self.assertTrue(self.main_subscription.database_ids)
        self.assertEqual(self.main_subscription._get_maintenance_loc_count(), 1000)
        self.main_subscription._create_invoices()._post()

        # start tour to activate the maintenance
        self.start_tour('/my/subscription', 'partner_activate_maintenance_yearly', login=self.referrer_user.login)
        self.assertTrue(self.main_subscription.maintenance_is_activated)

    @skip("test need rework for saas-15.3")
    def test_upsell(self):
        """ Upsell of new loc """

        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
            'module_2': 1000,
        }
        self.send_ping(db_uuid, modules)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])

        self.assertEqual(len(database), 1, "You should find a database")
        self.assertEqual(database.maintenance_lines_count, 2000, "The extra modules line count should be 2000 lines")
        database.subscription_id = self.main_subscription
        modules['module_1'] = 1500
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(database.maintenance_lines_count, 2500, "The extra modules line count should be 2000 lines")

        self.assertEqual(self.maintenance_line.product_uom_qty, 25.0, "Should be ten 100/loc per year")
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal', "No need to upsell at this stage")
        self.maintenance_subscription.start_subscription()
        modules['module_1'] = 2000
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertTrue(self.maintenance_subscription.to_renew, "Should need to upsell")
        # To renew should be True
        # Trigger Upsell sale order
        upsell_so = self.upsell(self.maintenance_subscription)
        # Confirm SO
        upsell_so.action_confirm()
        modules['module_1'] = 2500
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertFalse(self.maintenance_subscription.to_renew,
            "Should not trigger an upsell since the last upsell SO is less then 30 days ago")

        upsell_so.date_order = datetime.date.today() - datetime.timedelta(days=40)
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertTrue(self.maintenance_subscription.to_renew,
            "Should trigger an upsell since the last upsell is more than 30d days ago")

    @skip("test need rework for saas-15.3")
    def test_cancel_upsell(self):
        """ Test Cancel Upsell order """
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
            'module_2': 1000,
        }
        self.send_ping(db_uuid, modules)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        database.subscription_id = self.main_subscription
        self.maintenance_subscription.action_confirm()
        self.maintenance_subscription._create_invoices()._post()
        modules['module_1'] = 1500
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'upsell', "need to upsell at this stage")
        # Trigger Upsell sale order
        self.upsell(self.maintenance_subscription)
        modules['module_1'] = 1000
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal', "No need to upsell at this stage")

    @skip("test need rework for saas-15.3")
    def test_valid_date_maintenance_monthly(self):
        """ Test different use case that change the valid date of the database
            for the legacy use case: Yearly main subscription and monthly maintenance subscription """
        today = datetime.date.today()
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
            'module_2': 1000,
        }
        self.send_ping(db_uuid, modules)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), today + datetime.timedelta(days=31),
            "The database is not link to any subscription and is new, the expiration is in 31 days")

        # Delete yearly mnt sub to create a monthly one
        self.maintenance_subscription.unlink()
        database.subscription_id = self.main_subscription

        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), self.main_subscription.next_invoice_date + datetime.timedelta(days=1),
            "There is no maintenance subscription, the valid date should be the end of the main subscription + 1 days")

        monthly_template = self.env.ref('openerp_enterprise.subscription_template_odoo_monthly')

        line = {
            'product_id': self.maintenance_product.id,
            'product_uom_qty': 20.0,
            'uom_id': monthly_template.get_product_uom(self.maintenance_product).id,
            'price_unit': 16,
        }
        # TODO fixme
        maintenance_sub = self.env['sale.subscription'].create({
            'code': '%s/MNT' % self.main_subscription.client_order_ref,
            'date_start': datetime.date.today(),
            'next_invoice_date': datetime.date.today(),
            'name': self.main_subscription.name,
            'template_id': monthly_template.id,
            'subscription_state': '3_progress',
            'partner_id': self.main_subscription.partner_id.id,
            'pricelist_id': self.main_subscription.pricelist_id.id,
            'company_id': self.main_subscription.company_id.id,
            'order_line': [(0, 0, line)],
        })
        self.main_subscription.maintenance_subscription_id = maintenance_sub
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), maintenance_sub.next_invoice_date + datetime.timedelta(days=16),
            "The maintenance Subscription has no token, the expiration is a the end of the maintenance subscription + 16 days")

        maintenance_sub.payment_token_id = self.token
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), self.main_subscription.next_invoice_date + datetime.timedelta(days=1),
            "The maintenance subscription is ok, the valid date should be the end of the main subscription + 1 days")

    @skip("test need rework for saas-15.3")
    def test_valid_date(self):
        """ Test different use case that change the valid date of the database """
        db_uuid = generate_uuid()
        modules = {'module_1': 1000}
        self.send_ping(db_uuid, modules, code=self.code)
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), self.main_subscription.next_invoice_date + datetime.timedelta(days=1),
            "The maintenance subscription is draft, the valid date should be the end of the main subscription + 1 days")
        self.maintenance_subscription.action_confirm()
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), self.maintenance_subscription.next_invoice_date + datetime.timedelta(days=16),
            "The maintenance subscription need to be renewed, the valid date should be the end of the maintenance subscription + 16 days")
        res = self.maintenance_subscription.prepare_renewal_order()
        order = self.env['sale.order'].browse(res['res_id'])
        order.action_confirm()
        database.update_valid_date()
        self.assertEqual(database.date_valid.date(), self.main_subscription.next_invoice_date + datetime.timedelta(days=1),
            "The maintenance subscription is valid and paid, the valid date should be the end of the main subscription + 1 days")
        modules['module_1'] = 2500
        res = self.send_ping(db_uuid, modules, code=self.code)
        self.assertEqual(database.date_valid.date(), datetime.date.today() + relativedelta(months=1, days=1),
            "The maintenance subscription need an upsell, the valid date should be today + 1 month and 1 day")

    def test_constraint_one_mnt_line(self):
        """ Check it's not allowed to have 2 maintenance line on a subscription """
        with self.assertRaises(UserError):
            line = {
                'product_id': self.maintenance_product.id,
                'product_uom': self.main_subscription.sale_order_template_id.get_product_uom(self.maintenance_product).id,
                'price_unit': 16,
            }
            self.main_subscription.write({
                'order_line': [(0, 0, line)]
            })

    def test_sync_field_to_maintenance(self):
        """ Check that referrer, partner_id, commission_plan_id are synchronized
            with the maintenance subscription """

        # LegacyUseCase: This test is only need for maintenance subscriptions
        # created before # saas-15.3, where the maintenance could be on a
        # separate subscription
        self._setup_separate_maintenance_sub(self.customer, self.quotation_template)

        new_customer = self.customer.copy()
        new_referrer = self.referrer.copy()
        new_plan = self.referrer.commission_plan_id.copy()
        self.main_subscription.partner_id = new_customer
        self.assertEqual(self.maintenance_subscription.partner_id, new_customer,
            "partner_id should be synced with mnt subscription")
        self.main_subscription.referrer_id = new_referrer
        self.assertEqual(self.maintenance_subscription.referrer_id, new_referrer,
            "referrer_id should be synced with mnt subscription")
        self.main_subscription.commission_plan_id = new_plan
        self.assertEqual(self.maintenance_subscription.commission_plan_id, new_plan,
            "commission_plan_id should be synced with mnt subscription")

    def test_no_sync_field_to_maintenance(self):
        """ Check that referrer, partner_id, commission_plan_id are not synchronized
            from the maintenance subscription to the main subscription"""

        # LegacyUseCase: This test is only need for maintenance subscriptions
        # created before # saas-15.3, where the maintenance could be on a
        # separate subscription
        self._setup_separate_maintenance_sub(self.customer, self.quotation_template)

        new_customer = self.customer.copy()
        new_referrer = self.referrer.copy()
        new_plan = self.referrer.commission_plan_id.copy()
        self.maintenance_subscription.partner_id = new_customer
        self.assertEqual(self.main_subscription.partner_id, self.customer,
            "partner_id should remain the same on the main subsctiption")
        self.maintenance_subscription.referrer_id = new_referrer
        self.assertEqual(self.main_subscription.referrer_id, self.referrer,
            "referrer_id should remain the same on the main subsctiption")
        self.maintenance_subscription.commission_plan_id = new_plan
        self.assertEqual(self.main_subscription.commission_plan_id, self.referrer.commission_plan_id,
            "commission_plan_id should remain the same on the main subsctiption")

    @skip("test need rework for ~~saas-15.3~~ saas-16.2")
    def test_maintenance_reactivation(self):
        """ Check force the status to progress and renew = False
        will not set the maintenance sub in to renew with next invoice
        date = Today """
        db_uuid = generate_uuid()
        modules = {
            'module_1': 1000,
        }
        self.send_ping(db_uuid, modules, code=self.code)
        next_invoice_date = datetime.date.today() + relativedelta(years=1)
        self.maintenance_subscription.next_invoice_date = next_invoice_date
        self.maintenance_subscription.write({
            'subscription_state': '3_progress',
            'to_renew': False,
        })
        self.assertEqual(self.maintenance_subscription.next_invoice_date, next_invoice_date,
            "Next invoice date should remain unchanged")

class MaintenanceYearlyAccessRight(MaintenanceCommon):
    quotation_template = MaintenanceCommon.qt_custom_yearly

    def init_so(self):
        super().init_so()
        self.so.order_line.filtered(lambda l: l.product_id == self.maintenance_product).unlink()

    def test_see_subscription_maintenance_activation(self):
        """ Check that user with See Subscription can activate the maintenance.
            User without this group should not be able to activate it
        """
        self.main_subscription.partner_id = self.no_see_sub_user.partner_id
        self.assertFalse(self.main_subscription._get_maintenance_line(), "Maintenance should not be activated")
        with self.assertRaises(UserError):
            self.main_subscription.with_user(self.no_see_sub_user).action_activate_maintenance()

        self.maintenance_subscription.partner_id = self.see_sub_user.partner_id
        self.main_subscription.with_user(self.see_sub_user).action_activate_maintenance()
        self.assertTrue(self.main_subscription._get_maintenance_line(), "Maintenance should be activated")
